(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(factory((global['t-rex'] = global['t-rex'] || {})));
}(this, (function (exports) { 'use strict';

var enabled = false;
var nullLogger = {
    log: function () { },
    error: function () { },
    warn: function () { },
};
// function internalEnableLogging(enable) {
// }
var enableLogging = function (enable) {
    enabled = enable;
};
function createLogger(prefix) {
    if (!enabled) {
        return nullLogger;
    }
    return {
        log: console.log.bind(console, prefix + ">"),
        error: console.error.bind(console, prefix + ">"),
        warn: console.warn.bind(console, prefix + ">"),
    };
}

var ROOT = "/";
var $$MODIFIED = "$$t-rex:modified";
var $$VERSION = "$$t-rex:version";
var logger$1 = createLogger("TransactionalObject");
var TransactionalObject = (function () {
    function TransactionalObject(initialState) {
        this.base = this.current = initialState;
        this.modified = [];
    }
    TransactionalObject.prototype.commit = function () {
        logger$1.log("commit");
        for (var _i = 0, _a = this.modified; _i < _a.length; _i++) {
            var obj = _a[_i];
            delete obj[$$MODIFIED];
        }
        this.modified = [];
        this.base = this.current;
    };
    TransactionalObject.prototype.rebase = function (newBase) {
        logger$1.log("rebase");
        this.current = this.internalRebase(this.current, this.base, newBase, "/");
        this.base = newBase;
    };
    TransactionalObject.prototype.appendPath = function (path, field) {
        if (path == "/") {
            return "/" + field;
        }
        return path + "/" + field;
    };
    TransactionalObject.prototype.internalRebase = function (current, base, newBase, path) {
        if (newBase == base) {
            return current;
        }
        if (newBase[$$VERSION] < base[$$VERSION]) {
            return current;
        }
        var newCurrent = current;
        for (var field in newBase) {
            if (field.startsWith("$$t-rex")) {
                continue;
            }
            if (base[field] === current[field]) {
                //
                //  We hold no change for this field
                //  Any new change is welcomed
                //
                newCurrent = this.setField(current, field, newBase[field]);
                continue;
            }
            if (typeof base[field] !== "object") {
                //
                //  There is a local change inside a value type
                //  If newBase has different value than this is a conflict
                //
                if (base[field] != newBase[field]) {
                    var fieldPath = this.appendPath(path, field);
                    logger$1.error("Concurrency error at path \"" + fieldPath + "\".", "base is", base[field], "while latest is", newBase[field]);
                    throw new Error("Concurrency error at " + fieldPath);
                }
            }
            else {
                //
                //  There is a local change inside a reference type
                //  If newBase is older than our data than we are OK
                //
                if (base[field] === newBase[field] || base[field][$$VERSION] > newBase[field][$$VERSION]) {
                    continue;
                }
                //
                //  newBase is same version or bigger
                //  This implies a chance that newBase holds new values that might be conflicting
                //
                var newVal = this.internalRebase(current[field], base[field], newBase[field], (path == "/" ? path + field : path + "/" + field));
                newCurrent = this.setField(current, field, newVal);
            }
        }
        return newCurrent;
    };
    TransactionalObject.prototype.getBase = function () {
        return this.base;
    };
    TransactionalObject.prototype.getCurrent = function () {
        return this.current;
    };
    TransactionalObject.prototype.getProperty = function (root, path) {
        if (path == ROOT) {
            return root;
        }
        var parts = path.split(".");
        for (var _i = 0, parts_1 = parts; _i < parts_1.length; _i++) {
            var part = parts_1[_i];
            root = root[part];
        }
        return root;
    };
    TransactionalObject.prototype.setProperty = function (path, changes) {
        var pathEntries = this.getPath(this.current, path);
        var lastEntry = pathEntries[pathEntries.length - 1];
        var parent = (lastEntry ? lastEntry.parent[lastEntry.field] : this.current);
        var newValue = this.merge(parent, changes);
        for (var i = pathEntries.length - 1; i >= 0; i--) {
            var entry = pathEntries[i];
            var parent_1 = entry.parent;
            var field = entry.field;
            newValue = this.setField(parent_1, field, newValue);
        }
        this.current = newValue;
    };
    TransactionalObject.prototype.setField = function (parent, field, value) {
        if (parent[field] === value) {
            return parent;
        }
        var newParent = this.clone(parent);
        newParent[field] = value;
        return newParent;
    };
    TransactionalObject.prototype.getPath = function (obj, path) {
        var entries = [];
        if (path == ROOT) {
            return entries;
        }
        var parent = obj;
        var fields = path.split(".");
        for (var _i = 0, fields_1 = fields; _i < fields_1.length; _i++) {
            var field = fields_1[_i];
            if (parent === null || parent === undefined) {
                break;
            }
            if (typeof parent != "object") {
                throw new Error("Invalid path: " + path);
            }
            entries.push({ parent: parent, field: field });
            parent = parent[field];
        }
        return entries;
    };
    TransactionalObject.prototype.clone = function (obj) {
        if (obj && obj[$$MODIFIED]) {
            //
            //  This object is already a clone
            //  No need to clone again
            //
            return obj;
        }
        var res = Object.assign({}, obj);
        res[$$MODIFIED] = true;
        this.modified.push(res);
        res[$$VERSION] = (res[$$VERSION] || 0) + 1;
        return res;
    };
    TransactionalObject.prototype.merge = function (obj, changes) {
        if (obj == changes) {
            return changes;
        }
        if (typeof changes != "object") {
            //
            //  Cannot merge a value into itself
            //
            return changes;
        }
        if (Array.isArray(obj)) {
            //
            //  Cannot merge arrays
            //
            return changes;
        }
        var newObj = this.clone(obj);
        for (var field in changes) {
            if (changes.hasOwnProperty(field)) {
                var newValue = changes[field];
                var oldValue = newObj[field];
                if (oldValue != newValue) {
                    newObj[field] = newValue;
                }
            }
        }
        return newObj;
    };
    return TransactionalObject;
}());
TransactionalObject.ROOT = ROOT;

if (typeof Zone === "undefined") {
    throw new Error("t-rex cannot execute without zone.js. Please ensure zone.js is loaded before t-rex");
}
var TransactionScope = (function () {
    function TransactionScope(appStore) {
        this.id = ++TransactionScope.nextTranId;
        this.logger = createLogger("TransactionScope(" + this.id + ")");
        this.updateCount = 0;
        this.appStore = appStore;
        this.tranState = new TransactionalObject(appStore.getState());
        this.committed = false;
        this.outerZone = Zone.current;
        this.logger.log("created");
    }
    TransactionScope.prototype.update = function (path, changes) {
        ++this.updateCount;
        this.logger.log("update", path, changes);
        this.ensureNotCommitted();
        this.tranState.setProperty(path, changes);
    };
    TransactionScope.prototype.getNewState = function () {
        return this.tranState.getCurrent();
    };
    TransactionScope.prototype.getOldState = function () {
        return this.tranState.getBase();
    };
    TransactionScope.prototype.commit = function () {
        var _this = this;
        this.ensureNotCommitted();
        var oldState = this.tranState.getBase();
        var newState = this.tranState.getCurrent();
        var currentState = this.appStore.getState();
        if (newState == currentState) {
            this.logger.log("Nothing new to commit. updateCount is", this.updateCount);
            return;
        }
        if (oldState != currentState) {
            //
            //  A parallel transaction has already modified the main store before us
            //  We only allow parallel changes at different branches. Else, "Concurrency error is raised"
            //
            this.tranState.rebase(currentState);
            oldState = this.tranState.getBase();
            newState = this.tranState.getCurrent();
        }
        this.committed = true;
        this.outerZone.run(function () {
            //
            //  Committing to appStore causes emitting of change event
            //  Subscribers must be notified outside of the transaction zone, else, any
            //  additional update will be considered as part of the already committed transaction
            //  and therefore will throw error
            //
            _this.appStore.commit(oldState, newState);
        });
        //
        //  Cleans management flags + switch between old and new
        //
        this.tranState.commit();
        this.logger.log("committed");
    };
    TransactionScope.current = function () {
        var tran = Zone.current.get("tran");
        return tran;
    };
    TransactionScope.runInsideTransaction = function (appStore, action) {
        function runAction(func, commit) {
            var retVal = func();
            if (retVal && retVal.then) {
                var promise = retVal;
                return promise.then(function (res) {
                    if (commit) {
                        tran.commit();
                    }
                    return res;
                });
            }
            else {
                if (commit) {
                    tran.commit();
                }
                return retVal;
            }
        }
        var tran = TransactionScope.current();
        if (tran) {
            tran.ensureNotCommitted();
            //
            //  This is a nested transaction
            //  No need to commit changes to app base
            //
            return runAction(action, false);
        }
        //
        //  This is a root transaction
        //  When completed need to commit to the appStore
        //
        tran = new TransactionScope(appStore);
        var spec = {
            name: "tran",
            properties: {
                "tran": tran,
            },
        };
        var zone = Zone.current.fork(spec);
        return zone.run(function () {
            var tran1 = TransactionScope.current();
            return runAction(action, true);
        });
    };
    TransactionScope.prototype.ensureNotCommitted = function () {
        if (this.committed) {
            throw new Error("Activity was already committed");
        }
    };
    return TransactionScope;
}());
TransactionScope.nextTranId = 0;

function P1(obj, key1) {
    if (!obj) {
        return undefined;
    }
    return obj[key1]; // Inferred type is T[K]
}

function P2(obj, key1, key2) {
    return P1(P1(obj, key1), key2);
}



// export const promisify = (function() {
//     if(typeof window != "undefined" && window.hasOwnProperty("angular")) {
//         const angular = window["angular"];
//         let injector;
//         let $q;
//
//         return function promisifyOnAngular1(value) {
//             injector = injector || angular.element(document.getElementsByTagName("body")).injector();
//             $q = $q || injector.get("$q");
//
//             if(value && value.then) {
//                 return value;
//             }
//             return $q.when(value);
//         }
//     }
//     else {
//         return function promisifyOnNonAngular1(value) {
//             if(value && value.then) {
//                 return value;
//             }
//
//             return Promise.resolve(value);
//         }
//     }
// })();
//

var PathResolver = (function () {
    function PathResolver(path) {
        this.path = path;
        if (this.path == ROOT) {
            this.parts = [];
            return;
        }
        this.parts = this.path.split(".");
    }
    PathResolver.create = function (path) {
        return new PathResolver(path);
    };
    PathResolver.resolve = function (obj, path) {
        return new PathResolver(path).get(obj);
    };
    PathResolver.prototype.get = function (obj) {
        if (this.parts.length == 0) {
            return obj;
        }
        var res = obj;
        for (var _i = 0, _a = this.parts; _i < _a.length; _i++) {
            var part = _a[_i];
            res = res[part];
        }
        return res;
    };
    PathResolver.prototype.set = function (obj, val) {
        for (var i = 0; i < this.parts.length - 1; i++) {
            obj = obj[this.parts[i]];
        }
        obj[this.parts[this.parts.length - 1]] = val;
    };
    return PathResolver;
}());

var logger = createLogger("ServiceStore");
var ServiceStore = (function () {
    function ServiceStore(metadata) {
        this.appStore = null;
        this.metadata = metadata;
        this.pathResolver = new PathResolver(this.metadata.path);
        this.listeners = [];
    }
    ServiceStore.prototype.onAppStoreInitialized = function (appStore) {
        var _this = this;
        this.appStore = appStore;
        appStore.subscribe(function (newAppState, oldAppState) {
            var newState = _this.pathResolver.get(newAppState);
            var oldState = _this.pathResolver.get(oldAppState);
            if (oldState != newState) {
                _this.emit(newState, oldState);
            }
        });
    };
    ServiceStore.prototype.subscribe = function (listener) {
        this.ensureInitialized();
        this.listeners.push(listener);
        //
        //  Always notify new subscribers of the existing data
        //
        var state = this.pathResolver.get(this.appStore.getState());
        listener(state, state);
    };
    ServiceStore.prototype.subscribeTo = function (path, listener) {
        this.ensureInitialized();
        var pathResolver = new PathResolver(path);
        this.listeners.push(function (s1, s2) {
            var c1 = pathResolver.get(s1);
            var c2 = pathResolver.get(s2);
            if (c1 != c2) {
                listener(c1, c2);
            }
        });
    };
    ServiceStore.prototype.subscribeTo1 = function (key1, listener) {
        this.ensureInitialized();
        this.listeners.push(function (s1, s2) {
            var c1 = P1(s1, key1);
            var c2 = P1(s2, key1);
            if (c1 != c2) {
                listener(c1, c2);
            }
        });
    };
    ServiceStore.prototype.subscribeTo2 = function (key1, key2, listener) {
        this.ensureInitialized();
        this.listeners.push(function (s1, s2) {
            var c1 = P2(s1, key1, key2);
            var c2 = P2(s2, key1, key2);
            if (c1 != c2) {
                listener(c1, c2);
            }
        });
    };
    ServiceStore.prototype.getAppStore = function () {
        this.ensureInitialized();
        return this.appStore;
    };
    ServiceStore.prototype.getMetadata = function () {
        return this.metadata;
    };
    ServiceStore.prototype.getState = function () {
        this.ensureInitialized();
        var tranScope = TransactionScope.current();
        var appState = (tranScope ? tranScope.getNewState() : this.appStore.getState());
        var state = this.pathResolver.get(appState);
        return state;
    };
    ServiceStore.prototype.update = function (changes) {
        logger.log("update", changes);
        this.ensureInitialized();
        var tranScope = TransactionScope.current();
        if (!tranScope) {
            throw new Error("No ambient transaction to update");
        }
        tranScope.update(this.metadata.path, changes);
        var state = this.pathResolver.get(tranScope.getNewState());
        logger.log("update new state is", state);
        return state;
    };
    ServiceStore.create = function (path, initialState) {
        return new ServiceStore({
            path: path,
            initialState: initialState,
        });
    };
    ServiceStore.prototype.emit = function (newState, oldState) {
        for (var _i = 0, _a = this.listeners; _i < _a.length; _i++) {
            var l = _a[_i];
            try {
                l(newState, oldState);
            }
            catch (err) {
                logger.error("Ignoring error during ServiceStore change event", err);
            }
        }
    };
    ServiceStore.prototype.ensureInitialized = function () {
        if (!this.appStore) {
            throw new Error("Store is not initialized. Did you call appStore.init ?");
        }
    };
    return ServiceStore;
}());

var logger$2 = createLogger("AppStore");
// export abstract class AppStore<StateT extends object> {
//     abstract init(stores: ServiceStore<any>[]);
//
//     abstract getBase(): StateT;
//
//     abstract subscribe(listener: (newState, oldState) => void);
//
//     abstract unsubscribe(listener: (newState, oldState) => void);
//
//     abstract commit(oldState, newState);
//
//     static createStore<StateT extends object>(): AppStore<StateT> {
//         return new AppStoreImpl<StateT>();
//     }
// }
//
//  Holds application base and allow subscribing to changes
//  Each ServiceStore register itself to it
//  Commit request is delegated from the ServiceStore to this appStore
//
var AppStore = (function () {
    function AppStore() {
        this.appState = {};
        this.listeners = [];
        this.stores = [];
    }
    AppStore.prototype.init = function (stores) {
        for (var _i = 0, stores_1 = stores; _i < stores_1.length; _i++) {
            var store = stores_1[_i];
            if (store.store instanceof ServiceStore) {
                this.registerStore(store.store);
            }
        }
        for (var _a = 0, stores_2 = stores; _a < stores_2.length; _a++) {
            var store = stores_2[_a];
            if (store.store instanceof ServiceStore) {
                store.store.onAppStoreInitialized(this);
            }
        }
        logger$2.log("Initial appState", this.appState);
    };
    AppStore.prototype.getState = function () {
        return this.appState;
    };
    AppStore.prototype.subscribe = function (listener) {
        this.listeners.push(listener);
    };
    AppStore.prototype.unsubscribe = function (listener) {
        var index = this.listeners.findIndex(function (l) { return l == listener; });
        if (index != -1) {
            this.listeners.splice(index, 1);
        }
    };
    AppStore.prototype.commit = function (oldState, newState) {
        if (newState == this.appState) {
            logger$2.log("Nothing new to commit");
            return;
        }
        if (oldState != this.appState) {
            throw new Error("Concurrency error");
        }
        logger$2.log("Changing state", oldState, " ==> ", newState);
        this.appState = newState;
        this.emit(oldState, newState);
        return this.appState;
    };
    AppStore.prototype.emit = function (oldState, newState) {
        for (var _i = 0, _a = this.listeners; _i < _a.length; _i++) {
            var l = _a[_i];
            try {
                l(this.appState, oldState);
            }
            catch (err) {
                logger$2.error("Ignoring error during AppStore change event", err);
            }
        }
    };
    AppStore.prototype.registerStore = function (store) {
        var metadata = store.getMetadata();
        var conflict = this.findConflictingPath(metadata.path);
        if (conflict) {
            throw new Error("Service with path: " + metadata.path + " conflicts with existing service with path: " + conflict.getMetadata().path);
        }
        this.stores.push(store);
        if (store.getMetadata().path == ROOT) {
            this.appState = store.getMetadata().initialState;
            return;
        }
        PathResolver.create(metadata.path).set(this.appState, metadata.initialState);
    };
    AppStore.prototype.findConflictingPath = function (path) {
        if (path == ROOT) {
            return this.stores.find(function (s) { return s.getMetadata().path == ROOT; });
        }
        for (var _i = 0, _a = this.stores; _i < _a.length; _i++) {
            var store = _a[_i];
            if (path.indexOf(store.getMetadata().path) == 0) {
                return store;
            }
        }
        return null;
    };
    return AppStore;
}());

var logger$3 = createLogger("decorators");
function transaction(store, func) {
    var appStore = getAppStoreFromStore(store);
    return TransactionScope.runInsideTransaction(appStore, function () {
        return func();
    });
}
function getAppStoreFromStore(store) {
    if (store instanceof AppStore) {
        return store;
    }
    if (store instanceof ServiceStore) {
        return store.getAppStore();
    }
    throw new Error("Unexpected store value. Should be of type AppStore | ServiceStore");
}
function getAppStoreFromService(service) {
    var appStore;
    if (service.store) {
        if (!(service.store instanceof ServiceStore)) {
            throw new Error("Unexpected store value. Should be of type ServiceStore");
        }
        return service.store.getAppStore();
    }
    else if (service.appStore) {
        if (!(service.appStore instanceof AppStore)) {
            throw new Error("Unexpected appStore value. Should be of type AppStore");
        }
        return service.appStore;
    }
    throw new Error("No store/appStore field was found for service instance");
}

function getOptionOrDefault(options, propName, defValue) {
    var propVal;
    if (options && options.hasOwnProperty(propName)) {
        propVal = options[propName];
    }
    else {
        propVal = defValue;
    }
    return propVal;
}
function Activity(options) {
    return function (target, propertyKey, descriptor) {
        var method = descriptor.value;
        descriptor.value = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            var service = this;
            var appStore = getAppStoreFromService(service);
            var beginTransaction = getOptionOrDefault(options, "beginTransaction", true);
            if (beginTransaction) {
                return TransactionScope.runInsideTransaction(appStore, function () {
                    return method.apply(service, args);
                });
            }
            else {
                return method.apply(service, args);
            }
        };
        return descriptor;
    };
}

exports.ServiceStore = ServiceStore;
exports.AppStore = AppStore;
exports.Activity = Activity;
exports.transaction = transaction;
exports.TransactionScope = TransactionScope;
exports.enableLogging = enableLogging;

Object.defineProperty(exports, '__esModule', { value: true });

})));
